import json


class InvalidScheduleOutputFormat(Exception):
    """Raised if the scheduled output is on an invalid format."""

    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return json.dumps(self.errors, indent=4, sort_keys=True)


class DataframeMissingTimestampColumn(Exception):
    """Raised if trying to convert a dataframe without a timestamp column to scheduled output format."""

    pass


class DuplicateAliasInScheduledOutput(Exception):
    """Raised when an alias is passed more than once when converting to scheduled output format."""

    pass



import json


class SpecValidationError(Exception):
    """Raised if a data spec is invalid.

    Args:
        errors (Dict): A dictionary describing which fields are invalid and why.
    """

    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return json.dumps(self.errors, indent=4, sort_keys=True)



import json

from cognite.model_hosting.data_fetcher.exceptions import DataFetcherHttpError


def _status_is_valid(status_code: int) -> bool:
    return status_code < 400


def _raise_HTTP_error(code, x_request_id, response_body):
    extra = {}
    try:
        if isinstance(response_body, str):
            response_body = json.loads(response_body)
        error = response_body["error"]
        msg = error["message"]
        extra = error.get("extra")
    except:
        msg = response_body
    raise DataFetcherHttpError(msg, code, x_request_id, extra=extra)



import io
from typing import Dict, List, Union
from urllib.parse import quote

import pandas as pd

import cognite.model_hosting._cognite_model_hosting_common.utils as utils
from cognite.model_hosting.data_fetcher._client.api_client import ApiClient

DATAPOINTS_LIMIT = 100000
DATAPOINTS_LIMIT_AGGREGATES = 10000


class CdpClient(ApiClient):
    def get_datapoints_frame_single(
        self,
        id: int,
        start: int,
        end: int,
        aggregate: str = None,
        granularity: str = None,
        include_outside_points: bool = False,
    ) -> pd.DataFrame:
        limit = DATAPOINTS_LIMIT_AGGREGATES if aggregate else DATAPOINTS_LIMIT
        params = {
            "aggregates": aggregate,
            "granularity": granularity,
            "limit": limit,
            "start": start,
            "end": end,
            "includeOutsidePoints": include_outside_points,
        }
        ts = self.get_time_series_by_id([id])[0]
        url = "/timeseries/data/{}".format(quote(ts["name"], safe=""))
        datapoints = []
        while (not datapoints or len(datapoints[-1]) == limit) and params["end"] > params["start"]:
            res = self.get(url, params=params)
            res = res.json()["data"]["items"][0]["datapoints"]
            if not res:
                break
            datapoints.append(res)
            latest_timestamp = int(datapoints[-1][-1]["timestamp"])
            params["start"] = latest_timestamp + (utils.granularity_to_ms(granularity) if granularity else 1)
        dps = []
        [dps.extend(el) for el in datapoints]
        df = pd.DataFrame(dps)
        if include_outside_points:
            df.drop_duplicates(inplace=True)
        return df

    def get_datapoints_frame(
        self, time_series: List[Dict[str, Union[int, str]]], granularity: str, start: int, end: int
    ) -> pd.DataFrame:
        limit = DATAPOINTS_LIMIT // len(time_series)

        ts_ids = [ts["id"] for ts in time_series]
        ts_names = {ts["id"]: ts["name"] for ts in (self.get_time_series_by_id(ts_ids))}

        time_series_by_name = [{"name": ts_names[ts["id"]], "aggregates": [ts["aggregate"]]} for ts in time_series]
        body = {"items": time_series_by_name, "granularity": granularity, "start": start, "end": end, "limit": limit}
        url = "/timeseries/dataframe"

        dataframes = []
        while (not dataframes or dataframes[-1].shape[0] == limit) and body["end"] > body["start"]:
            res = self.post(url=url, json=body, headers={"accept": "text/csv"})
            dataframes.append(pd.read_csv(io.StringIO(res.text)))
            if dataframes[-1].empty:
                break
            latest_timestamp = int(dataframes[-1].iloc[-1, 0])
            body["start"] = latest_timestamp + utils.granularity_to_ms(granularity)
        return pd.concat(dataframes).reset_index(drop=True)

    def get_time_series_by_id(self, ids: List[int]) -> List:
        url = "/timeseries/byids"
        body = {"items": list(set(ids))}
        return (self.post(url, json=body, api_version="0.6")).json()["data"]["items"]

    def download_file(self, id: int, target_path: str, chunk_size: int = 2 ** 21) -> None:
        url = "/files/{}/downloadlink".format(id)
        download_url = self.get(url=url).json()["data"]

        with self._requests_session.get(download_url, stream=True) as r:
            with open(target_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)

    def download_file_to_memory(self, id) -> bytes:
        url = "/files/{}/downloadlink".format(id)
        download_url = (self.get(url=url)).json()["data"]

        with self._requests_session.get(download_url) as response:
            return response.content



import gzip
import json
import logging
import os
import sys
from typing import Any, Dict, Optional, Union

import requests
from requests import Response
from requests.adapters import HTTPAdapter
from urllib3 import Retry

import cognite.model_hosting._cognite_model_hosting_common.version
from cognite.model_hosting.data_fetcher._client.utils import _status_is_valid
from cognite.model_hosting.data_fetcher.exceptions import ApiKeyError, DataFetcherHttpError

DEFAULT_BASE_URL = "https://api.cognitedata.com"
DEFAULT_NUM_OF_RETRIES = 3
DEFAULT_TIMEOUT = 10

HTTP_STATUS_CODES_TO_RETRY = [429, 500, 502, 503]
HTTP_METHODS_TO_RETRY = ["GET", "DELETE"]

log = logging.getLogger("data-fetcher")


def _init_requests_session():
    session = requests.Session()
    retry = Retry(
        total=DEFAULT_NUM_OF_RETRIES,
        backoff_factor=0.3,
        status_forcelist=HTTP_STATUS_CODES_TO_RETRY,
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


_REQUESTS_SESSION = _init_requests_session()


def _raise_API_error(res: Response):
    x_request_id = res.headers.get("X-Request-Id")
    code = res.status_code
    extra = {}
    try:
        error = res.json()["error"]
        if isinstance(error, str):
            msg = error
        else:
            msg = error["message"]
            extra = error.get("extra")
    except:
        msg = res.content

    log.error("HTTP Error %s: %s", code, msg, extra={"X-Request-ID": x_request_id, "extra": extra})
    raise DataFetcherHttpError(msg, code, x_request_id, extra=extra)


class ApiClient:
    def __init__(self, api_key: str = None, project: str = None, base_url: str = None):
        thread_local_api_key, thread_local_project = self._get_thread_local_credentials()
        environment_api_key = os.getenv("COGNITE_API_KEY")
        environment_base_url = os.getenv("COGNITE_BASE_URL")

        self._requests_session = _REQUESTS_SESSION
        self._api_key = api_key or thread_local_api_key or environment_api_key
        self._base_url = base_url or environment_base_url or DEFAULT_BASE_URL
        self._headers = {
            "api-key": self._api_key,
            "content-type": "application/json",
            "accept": "application/json",
            "User-Agent": "CogniteDataFetcher/{}".format(
                cognite.model_hosting._cognite_model_hosting_common.version.__version__
            ),
        }
        self._project = project or thread_local_project or self._get_project(self._api_key)
        self._base_url_v0_5 = self._base_url + "/api/0.5/projects/{}".format(self._project)
        self._base_url_v0_6 = self._base_url + "/api/0.6/projects/{}".format(self._project)

    def get(
        self,
        url: str,
        api_version: Optional[str] = "0.5",
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        return self._do_request("GET", url, api_version=api_version, headers=headers, params=params)

    def post(
        self,
        url: str,
        api_version: Optional[str] = "0.5",
        json: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        return self._do_request("POST", url, api_version=api_version, headers=headers, params=params, data=json)

    def delete(
        self,
        url: str,
        api_version: Optional[str] = "0.5",
        json: Optional[Any] = None,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        return self._do_request("DELETE", url, api_version=api_version, headers=headers, params=params, data=json)

    def _do_request(
        self,
        method: str,
        url: str,
        api_version: Optional[str] = "0.5",
        data: Optional[Union[str, bytes]] = None,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        full_url = self._get_full_url(url, api_version)
        headers = {**self._headers, **(headers or {})}
        if data:
            headers["Content-Encoding"] = "gzip"
            data = gzip.compress(json.dumps(data).encode())
        response = requests.request(
            method, full_url, headers=headers, params=params, data=data, timeout=DEFAULT_TIMEOUT
        )
        if not _status_is_valid(response.status_code):
            _raise_API_error(response)
        return response

    def _get_project(self, api_key: str) -> str:
        if api_key is None:
            raise ApiKeyError("No API key was specified")
        response = self.get("/login/status", api_version=None).json()
        if not response["data"]["loggedIn"]:
            raise ApiKeyError("Invalid API Key")
        return response["data"]["project"]

    def _get_full_url(self, url, api_version: str = None):
        if api_version == "0.5":
            return self._base_url_v0_5 + url
        elif api_version == "0.6":
            return self._base_url_v0_6 + url
        return self._base_url + url

    def _get_thread_local_credentials(self):
        thread_local_api_key = None
        thread_local_project = None
        if "cognite._thread_local" in sys.modules:
            from cognite._thread_local import credentials

            thread_local_api_key = getattr(credentials, "api_key", None)
            thread_local_project = getattr(credentials, "project", None)
        return thread_local_api_key, thread_local_project



from .data_spec import (
    DataSpec,
    FileSpec,
    ScheduleDataSpec,
    ScheduleInputSpec,
    ScheduleInputTimeSeriesSpec,
    ScheduleOutputSpec,
    ScheduleOutputTimeSeriesSpec,
    TimeSeriesSpec,
)



import re

from setuptools import find_packages, setup

version = re.search(
    '^__version__\s*=\s*"(.*)"', open("cognite/model_hosting/_cognite_model_hosting_common/version.py").read(), re.M
).group(1)

setup(
    name="cognite-model-hosting",
    version=version,
    description="Utilities Cognite's model hosting environment",
    url="",  # TODO
    author="Nils Barlaug",
    author_email="nils.barlaug@cognite.com",
    packages=["cognite.model_hosting." + p for p in find_packages(where="cognite/model_hosting")],
    install_requires=["pandas", "marshmallow==3.0.0rc4", "requests>=2.21.0,<3.0.0"],
    python_requires=">=3.5",
)
